"""
Display uptime for workers in running Storm topologies.
"""

from __future__ import absolute_import, print_function

from pkg_resources import parse_version
from prettytable import PrettyTable

from .common import add_environment
from ..util import get_ui_json, storm_lib_version


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('worker_uptime',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)


def display_worker_uptime(env_name):
    topology_summary = '/api/v1/topology/summary'
    topology_detail = '/api/v1/topology/{topology}'
    component = '/api/v1/topology/{topology}/component/{component}'
    topo_summary_json = get_ui_json(env_name, topology_summary)
    topology_ids = [x['id'] for x in topo_summary_json['topologies']]
    worker_stats = []

    for topology in topology_ids:
        topology_detail_json = get_ui_json(env_name,
                                           topology_detail.format(topology=topology))
        spouts = [x['spoutId'] for x in topology_detail_json['spouts']]
        bolts = [x['boltId'] for x in topology_detail_json['bolts']]
        for comp in spouts + bolts:
            comp_detail = get_ui_json(env_name,
                                      component.format(topology=topology,
                                                       component=comp))
            worker_stats += [(worker['host'], worker['id'], worker['uptime'],
                              worker['workerLogLink']) for worker in
                             comp_detail['executorStats']]
    worker_stats = sorted(set(worker_stats))

    print("# Worker Stats")
    table = PrettyTable(["Host", "Worker ID", "Uptime", "Log URL"])
    table.align = 'l'
    table.align['Uptime'] = 'r'
    for row in worker_stats:
        table.add_row(row)
    print(table)
    print()


def main(args):
    """ Display uptime for Storm workers. """
    storm_version = storm_lib_version()
    if storm_version >= parse_version('0.9.2-incubating'):
        display_worker_uptime(args.environment)
    else:
        print("ERROR: Storm {0} does not support this command."
              .format(storm_version))



"""
Create or update a virtualenv on the Storm workers.  This will be done
automatically upon submit, but this command is provided to help with testing
and debugging.
"""

from __future__ import absolute_import, print_function, unicode_literals

import os
from io import open

from fabric.api import env, execute, parallel, prefix, put, puts, run
from fabric.contrib.files import exists

from .common import add_environment, add_name
from ..util import activate_env, die, get_config, get_topology_definition


@parallel
def _create_or_update_virtualenv(virtualenv_root,
                                 virtualenv_name,
                                 requirements_file,
                                 virtualenv_flags=None):
    virtualenv_path = os.path.join(virtualenv_root, virtualenv_name)
    if not exists(virtualenv_path):
        if virtualenv_flags is None:
            virtualenv_flags = ''
        puts("virtualenv not found in {}, creating one.".format(virtualenv_root))
        run("virtualenv {} {}".format(virtualenv_path, virtualenv_flags))

    puts("Uploading requirements.txt to temporary file.")
    tmpfile = run("mktemp /tmp/streamparse_requirements-XXXXXXXXX.txt")
    put(requirements_file, tmpfile)

    puts("Updating virtualenv: {}".format(virtualenv_name))
    cmd = "source {}".format(os.path.join(virtualenv_path, 'bin/activate'))
    with prefix(cmd):
        run("pip install -r {} --exists-action w".format(tmpfile))

    run("rm {}".format(tmpfile))


def create_or_update_virtualenvs(env_name, topology_name, requirements_file,
                                 virtualenv_flags=None):
    """Create or update virtualenvs on remote servers.

    Assumes that virtualenv is on the path of the remote server(s).

    :param env_name: the name of the environment in config.json.
    :param topology_name: the name of the topology (and virtualenv).
    :param requirements_file: path to the requirements.txt file to use
                              to update/install this virtualenv.
    """
    activate_env(env_name)
    # Check to ensure streamparse is in requirements
    with open(requirements_file, "r") as fp:
        found_streamparse = False
        for line in fp:
            if "streamparse" in line:
                found_streamparse = True
                break

        if not found_streamparse:
            die("Could not find streamparse in your requirements file ({}). "
                "streamparse is required for all topologies."
                .format(requirements_file))

    execute(_create_or_update_virtualenv, env.virtualenv_root, topology_name,
            requirements_file, virtualenv_flags=virtualenv_flags,
            hosts=env.storm_workers)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('update_virtualenv',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    add_name(subparser)


def main(args):
    """ Create or update a virtualenv on Storm workers. """
    name = get_topology_definition(args.name)[0]
    config = get_config()
    config["virtualenv_specs"] = config["virtualenv_specs"].rstrip("/")
    create_or_update_virtualenvs(args.environment, name,
                                 "{}/{}.txt".format(config["virtualenv_specs"],
                                                    name))



"""Base primititve classes for working with Storm."""
from __future__ import absolute_import, print_function, unicode_literals

import io
import logging
import os
import signal
import sys
from collections import deque, namedtuple
from logging.handlers import RotatingFileHandler
from os.path import join
from threading import RLock
from traceback import format_exc

import simplejson as json


# Support for Storm Log levels as per STORM-414
_STORM_LOG_TRACE = 0
_STORM_LOG_DEBUG = 1
_STORM_LOG_INFO = 2
_STORM_LOG_WARN = 3
_STORM_LOG_ERROR = 4
_STORM_LOG_LEVELS = {
    'trace': _STORM_LOG_TRACE,
    'debug': _STORM_LOG_DEBUG,
    'info': _STORM_LOG_INFO,
    'warn': _STORM_LOG_WARN,
    'warning': _STORM_LOG_WARN,
    'error': _STORM_LOG_ERROR,
    'critical': _STORM_LOG_ERROR
}
_PYTHON_LOG_LEVELS = {
    'critical': logging.CRITICAL,
    'error': logging.ERROR,
    'warning': logging.WARNING,
    'warn': logging.WARNING,
    'info': logging.INFO,
    'debug': logging.DEBUG,
    'trace': logging.DEBUG
}


log = logging.getLogger(__name__)


def remote_pdb_handler(signum, frame):
    """ Handler to drop us into a remote debugger upon receiving SIGUSR1 """
    try:
        from remote_pdb import RemotePdb
        rdb = RemotePdb(host='127.0.0.1', port=0)
        rdb.set_trace(frame=frame)
    except ImportError:
        log.warning('remote_pdb unavailable.  Please install remote_pdb to '
                    'allow remote debugging.')
    # Restore signal handler for later
    signal.signal(signum, remote_pdb_handler)


class StormHandler(logging.Handler):
    """Handler that will send messages back to Storm."""

    def __init__(self, stream=None):
        """ Initialize handler """
        if stream is None:
            stream = sys.stdout
        super(StormHandler, self).__init__()
        self._component = Component(output_stream=stream)

    def emit(self, record):
        """
        Emit a record.

        If a formatter is specified, it is used to format the record.
        If exception information is present, it is formatted using
        traceback.print_exception and sent to Storm.
        """
        try:
            msg = self.format(record)
            level = _STORM_LOG_LEVELS.get(record.levelname.lower(),
                                          _STORM_LOG_INFO)
            self._component.send_message({'command': 'log', 'msg': str(msg),
                                          'level': level})
        except Exception:
            self.handleError(record)


class LogStream(object):
    """Object that implements enough of the Python stream API to be used as
    sys.stdout. Messages are written to the Python logger.
    """
    def __init__(self, logger):
        self.logger = logger

    def write(self, message):
        if message.strip() == "":
            return  # skip blank lines

        try:
            self.logger.info(message)
        except:
            # There's been an issue somewhere in the logging sub-system
            # so we'll put stderr and stdout back to their originals and
            # raise the exception which will cause Storm to choke
            sys.stdout = sys.__stdout__
            raise

    def flush(self):
        """No-op method to prevent crashes when someone does
        sys.stdout.flush.
        """
        pass


Tuple = namedtuple('Tuple', 'id component stream task values')
"""Storm's primitive data type passed around via streams.

:ivar id: the ID of the Tuple.
:type id: str
:ivar component: component that the Tuple was generated from.
:type component: str
:ivar stream: the stream that the Tuple was emitted into.
:type stream: str
:ivar task: the task the Tuple was generated from.
:type task: int
:ivar values: the payload of the Tuple where data is stored.
:type values: list
"""


class Component(object):
    """Base class for spouts and bolts which contains class methods for
    logging messages back to the Storm worker process.


    :ivar input_stream: The ``file``-like object to use to retrieve commands
                        from Storm.  Defaults to ``sys.stdin``.
    :ivar output_stream: The ``file``-like object to send messages to Storm with.
                         Defaults to ``sys.stdout``.
    :ivar topology_name: The name of the topology sent by Storm in the initial
                         handshake.
    :ivar task_id: The numerical task ID for this component, as sent by Storm in
                   the initial handshake.
    :ivar component_name: The name of this component, as sent by Storm in the
                          initial handshake.
    :ivar debug: A ``bool`` indicating whether or not Storm is running in debug
                 mode.  Specified by the `topology.debug` Storm setting.
    :ivar storm_conf: A ``dict`` containing the configuration values sent by
                      Storm in the initial handshake with this component.
    :ivar context: The context of where this component is in the topology.  See
                   `the Storm Multi-Lang protocol documentation <https://storm.apache.org/documentation/Multilang-protocol.html>`__
                   for details.
    :ivar pid: An ``int`` indicating the process ID of this component as
               retrieved by ``os.getpid()``.
    :ivar logger: A logger to use with this component.

                  .. note::
                    Using ``Component.logger`` combined with the
                    :class:`streamparse.storm.component.StormHandler` handler is
                    the recommended way for logging messages from your
                    component. If you use ``Component.log`` instead, the logging
                    messages will *always* be sent to Storm, even if they are
                    ``debug`` level messages and you are running in production.
                    Using :class:`streamparse.storm.component.StormHandler`
                    ensures that you will instead have your logging messages
                    filtered on the Python side and only have the messages you
                    actually want logged serialized and sent to Storm.
    """


    def __init__(self, input_stream=sys.stdin, output_stream=sys.stdout,
                 rdb_signal=signal.SIGUSR1):
        # Ensure we don't fall back on the platform-dependent encoding and
        # always use UTF-8
        self.input_stream = self._wrap_stream(input_stream)
        self.output_stream = self._wrap_stream(output_stream)
        self.topology_name = None
        self.task_id = None
        self.component_name = None
        self.debug = None
        self.storm_conf = None
        self.context = None
        self.pid = os.getpid()
        self.logger = None
        # pending commands/Tuples we read while trying to read task IDs
        self._pending_commands = deque()
        # pending task IDs we read while trying to read commands/Tuples
        self._pending_task_ids = deque()
        self._reader_lock = RLock()
        self._writer_lock = RLock()
        # Setup remote pdb handler if asked to
        if rdb_signal is not None:
            signal.signal(rdb_signal, remote_pdb_handler)

    @staticmethod
    def is_heartbeat(tup):
        """ :returns: Whether or not the given Tuple is a heartbeat """
        return tup.task == -1 and tup.stream == '__heartbeat'

    @staticmethod
    def _wrap_stream(stream):
        """Returns a TextIOWrapper around the given stream that handles UTF-8
        encoding/decoding.
        """
        if hasattr(stream, 'buffer'):
            return io.TextIOWrapper(stream.buffer, encoding='utf-8')
        elif hasattr(stream, 'readable'):
            return io.TextIOWrapper(stream, encoding='utf-8')
        # Python 2.x stdin and stdout are just files
        else:
            return io.open(stream.fileno(), mode=stream.mode, encoding='utf-8')

    def _setup_component(self, storm_conf, context):
        """Add helpful instance variables to component after initial handshake
        with Storm.  Also configure logging.
        """
        self.topology_name = storm_conf.get('topology.name', '')
        self.task_id = context.get('taskid', '')
        self.component_name = context.get('componentid')
        # If using Storm before 0.10.0 componentid is not available
        if self.component_name is None:
            self.component_name = context.get('task->component', {})\
                                         .get(str(self.task_id), '')
        self.debug = storm_conf.get("topology.debug", False)
        self.storm_conf = storm_conf
        self.context = context

        # Set up logging
        self.logger = logging.getLogger('.'.join((__name__,
                                                  self.component_name)))
        log_path = self.storm_conf.get('streamparse.log.path')
        if log_path:
            root_log = logging.getLogger()
            max_bytes = self.storm_conf.get('streamparse.log.max_bytes',
                                            1000000)  # 1 MB
            backup_count = self.storm_conf.get('streamparse.log.backup_count',
                                               10)
            log_file = join(log_path,
                            ('streamparse_{topology_name}_{component_name}'
                             '_{task_id}_{pid}.log'
                             .format(topology_name=self.topology_name,
                                     component_name=self.component_name,
                                     task_id=self.task_id,
                                     pid=self.pid)))
            handler = RotatingFileHandler(log_file, maxBytes=max_bytes,
                                          backupCount=backup_count)
            formatter = logging.Formatter('%(asctime)s - %(name)s - '
                                          '%(levelname)s - %(message)s')
            handler.setFormatter(formatter)
            root_log.addHandler(handler)
            log_level = self.storm_conf.get('streamparse.log.level', 'info')
            log_level = _PYTHON_LOG_LEVELS.get(log_level, logging.INFO)
            if self.debug:
                # potentially override logging that was provided if
                # topology.debug was set to true
                log_level = logging.DEBUG
            root_log.setLevel(log_level)
        else:
            self.send_message({'command': 'log',
                               'msg': ('WARNING: streamparse logging is not '
                                       'configured. Please set streamparse.log.'
                                       'path in your config.json.')})
        # Redirect stdout to ensure that print statements/functions
        # won't disrupt the multilang protocol
        sys.stdout = LogStream(logging.getLogger('streamparse.stdout'))

    def read_message(self):
        """Read a message from Storm, reconstruct newlines appropriately.

        All of Storm's messages (for either bolts or spouts) should be of the
        form::

            '<command or task_id form prior emit>\\nend\\n'

        Command example, an incoming Tuple to a bolt::

            '{ "id": "-6955786537413359385",  "comp": "1", "stream": "1", "task": 9, "tuple": ["snow white and the seven dwarfs", "field2", 3]}\\nend\\n'

        Command example for a spout to emit its next Tuple::

            '{"command": "next"}\\nend\\n'

        Example, the task IDs a prior emit was sent to::

            '[12, 22, 24]\\nend\\n'

        The edge case of where we read ``''`` from ``input_stream`` indicating
        EOF, usually means that communication with the supervisor has been
        severed.
        """
        msg = ""
        num_blank_lines = 0
        while True:
            # readline will return trailing \n so that output is unambigious, we
            # should only have line == '' if we're at EOF
            with self._reader_lock:
                line = self.input_stream.readline()
            if line == 'end\n':
                break
            elif line == '':
                log.error("Received EOF while trying to read stdin from Storm, "
                           "pipe appears to be broken, exiting.")
                sys.exit(1)
            elif line == '\n':
                num_blank_lines += 1
                if num_blank_lines % 1000 == 0:
                    log.warn("While trying to read a command or pending task "
                             "ID, Storm has instead sent %s '\\n' messages.",
                             num_blank_lines)
                continue

            msg = '{}{}\n'.format(msg, line[0:-1])

        try:
            return json.loads(msg)
        except Exception:
            log.error("JSON decode error for message: %r", msg, exc_info=True)
            raise

    def read_task_ids(self):
        if self._pending_task_ids:
            return self._pending_task_ids.popleft()
        else:
            msg = self.read_message()
            while not isinstance(msg, list):
                self._pending_commands.append(msg)
                msg = self.read_message()
            return msg

    def read_command(self):
        if self._pending_commands:
            return self._pending_commands.popleft()
        else:
            msg = self.read_message()
            while isinstance(msg, list):
                self._pending_task_ids.append(msg)
                msg = self.read_message()
            return msg

    def read_tuple(self):
        cmd = self.read_command()
        return Tuple(cmd['id'], cmd['comp'], cmd['stream'], cmd['task'],
                     cmd['tuple'])

    def read_handshake(self):
        """Read and process an initial handshake message from Storm."""
        msg = self.read_message()
        pid_dir, _conf, _context = msg['pidDir'], msg['conf'], msg['context']

        # Write a blank PID file out to the pidDir
        open(join(pid_dir, str(self.pid)), 'w').close()
        self.send_message({'pid': self.pid})

        return _conf, _context

    def send_message(self, message):
        """Send a message to Storm via stdout."""
        if not isinstance(message, dict):
            log.error("%s.%d attempted to send a non dict message to Storm: %r",
                       self.component_name, self.pid, message)
            return

        wrapped_msg = "{}\nend\n".format(json.dumps(message))

        with self._writer_lock:
            self.output_stream.flush()
            self.output_stream.write(wrapped_msg)
            self.output_stream.flush()

    def raise_exception(self, exception, tup=None):
        """Report an exception back to Storm via logging.

        :param exception: a Python exception.
        :param tup: a :class:`Tuple` object.
        """
        if tup:
            message = ('Python {exception_name} raised while processing Tuple '
                       '{tup!r}\n{traceback}')
        else:
            message = 'Python {exception_name} raised\n{traceback}'
        message = message.format(exception_name=exception.__class__.__name__,
                                 tup=tup,
                                 traceback=format_exc())
        self.send_message({'command': 'error', 'msg': str(message)})
        self.send_message({'command': 'sync'})  # sync up right away

    def log(self, message, level=None):
        """Log a message to Storm optionally providing a logging level.

        :param message: the log message to send to Storm.
        :type message: str
        :param level: the logging level that Storm should use when writing the
                      ``message``. Can be one of: trace, debug, info, warn, or
                      error (default: ``info``).
        :type level: str

        .. warning::

          This will send your message to Storm regardless of what level you
          specify.  In almost all cases, you are better of using
          ``Component.logger`` with a
          :class:`streamparse.storm.component.StormHandler`, because the
          filtering will happen on the Python side (instead of on the Java side
          after taking the time to serialize your message and send it to Storm).
        """
        level = _STORM_LOG_LEVELS.get(level, _STORM_LOG_INFO)
        self.send_message({'command': 'log', 'msg': str(message),
                           'level': level})

    def emit(self, tup, tup_id=None, stream=None, anchors=None,
             direct_task=None, need_task_ids=True):
        """Emit a new Tuple to a stream.

        :param tup: the Tuple payload to send to Storm, should contain only
                    JSON-serializable data.
        :type tup: :class:`list` or :class:`streamparse.storm.component.Tuple`
        :param tup_id: the ID for the Tuple. If omitted by a
                       :class:`streamparse.storm.spout.Spout`, this emit will be
                       unreliable.
        :type tup_id: str
        :param stream: the ID of the stream to emit this Tuple to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param anchors: IDs the Tuples (or
                        :class:`streamparse.storm.component.Tuple` instances)
                        which the emitted Tuples should be anchored to. This is
                        only passed by :class:`streamparse.storm.bolt.Bolt`.
        :type anchors: list
        :param direct_task: the task to send the Tuple to.
        :type direct_task: int
        :param need_task_ids: indicate whether or not you'd like the task IDs
                              the Tuple was emitted (default: ``True``).
        :type need_task_ids: bool

        :returns: a ``list`` of task IDs that the Tuple was sent to. Note that
                  when specifying direct_task, this will be equal to
                  ``[direct_task]``. If you specify ``need_task_ids=False``,
                  this function will return ``None``.
        """
        if not isinstance(tup, (list, tuple)):
            raise TypeError('All Tuples must be either lists or tuples, '
                            'received {!r} instead.'.format(type(tup)))

        msg = {'command': 'emit', 'tuple': tup}

        if anchors is not None:
            msg['anchors'] = anchors
        if tup_id is not None:
            msg['id'] = tup_id
        if stream is not None:
            msg['stream'] = stream
        if direct_task is not None:
            msg['task'] = direct_task

        if need_task_ids is False:
            # only need to send on False, Storm's default is True
            msg['need_task_ids'] = need_task_ids

        # Use both locks so we ensure send_message and read_task_ids are for
        # same emit
        with self._reader_lock, self._writer_lock:
            # Message encoding will convert both list and tuple to a JSON array.
            self.send_message(msg)

            if need_task_ids is True:
                downstream_task_ids = [direct_task] if direct_task is not None \
                                      else self.read_task_ids()
                return downstream_task_ids
            else:
                return None

    def run(self):
        """Main run loop for all components.

        Performs initial handshake with Storm and reads Tuples handing them off
        to subclasses.  Any exceptions are caught and logged back to Storm
        prior to the Python process exiting.

        .. warning::

            Subclasses should **not** override this method.
        """
        storm_conf, context = self.read_handshake()
        self._setup_component(storm_conf, context)

        try:
            self.initialize(storm_conf, context)
            while True:
                self._run()
        except Exception as e:
            self._handle_run_exception(e)
            sys.exit(1)

    def _handle_run_exception(self, exc):
        """Process an exception encountered while running the ``run()`` loop.

        Called right before program exits.
        """
        log_msg = "Exception in {}.run()".format(self.__class__.__name__)
        log.error(log_msg, exc_info=True)
        self.raise_exception(exc)



"""
List the currently running Storm topologies.
"""

from __future__ import absolute_import

from invoke import run

from .common import add_environment
from ..util import (get_env_config, get_nimbus_for_env_config,
                        is_ssh_for_nimbus)
from ..contextmanagers import ssh_tunnel


def _list_topologies(host=None, port=None, run_args=None, run_kwargs=None):
    if run_args is None:
        run_args = []
    if run_kwargs is None:
        run_kwargs = {}
    run_kwargs['pty'] = True
    cmd = ["lein",
           "run -m streamparse.commands.list/-main"]
    if host:
        cmd.append("--host {}".format(host))
    if port:
        cmd.append("--port {}".format(port))
    return run(" ".join(cmd), *run_args, **run_kwargs)


def list_topologies(env_name):
    env_name, env_config = get_env_config(env_name)
    host, port = get_nimbus_for_env_config(env_config)

    if is_ssh_for_nimbus(env_config):
        with ssh_tunnel(env_config.get("user"), host, 6627, port):
            return _list_topologies()
    return _list_topologies(host=host, port=port)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('list',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)


def main(args):
    """ List the currently running Storm topologies """
    list_topologies(args.environment)



"""
Base Spout classes.
"""

from __future__ import absolute_import, print_function, unicode_literals

import itertools
import logging

from six.moves import zip

from .component import Component


log = logging.getLogger(__name__)


class Spout(Component):
    """Base class for all streamparse spouts.

    For more information on spouts, consult Storm's
    `Concepts documentation <http://storm.apache.org/documentation/Concepts.html>`_.
    """

    def initialize(self, storm_conf, context):
        """Called immediately after the initial handshake with Storm and before
        the main run loop. A good place to initialize connections to data
        sources.

        :param storm_conf: the Storm configuration for this spout. This is the
                           configuration provided to the topology, merged in
                           with cluster configuration on the worker node.
        :type storm_conf: dict
        :param context: information about the component's place within the
                        topology such as: task IDs, inputs, outputs etc.
        :type context: dict
        """
        pass

    def ack(self, tup_id):
        """Called when a bolt acknowledges a Tuple in the topology.

        :param tup_id: the ID of the Tuple that has been fully acknowledged in
                       the topology.
        :type tup_id: str
        """
        pass

    def fail(self, tup_id):
        """Called when a Tuple fails in the topology

        A spout can choose to emit the Tuple again or ignore the fail. The
        default is to ignore.

        :param tup_id: the ID of the Tuple that has failed in the topology
                       either due to a bolt calling ``fail()`` or a Tuple
                       timing out.
        :type tup_id: str
        """
        pass

    def next_tuple(self):
        """Implement this function to emit Tuples as necessary.

        This function should not block, or Storm will think the
        spout is dead. Instead, let it return and streamparse will
        send a noop to storm, which lets it know the spout is functioning.
        """
        raise NotImplementedError()

    def emit(self, tup, tup_id=None, stream=None, direct_task=None,
             need_task_ids=True):
        """Emit a spout Tuple message.

        :param tup: the Tuple to send to Storm, should contain only
                    JSON-serializable data.
        :type tup: list or tuple
        :param tup_id: the ID for the Tuple. Leave this blank for an
                       unreliable emit.
        :type tup_id: str
        :param stream: ID of the stream this Tuple should be emitted to.
                       Leave empty to emit to the default stream.
        :type stream: str
        :param direct_task: the task to send the Tuple to if performing a
                            direct emit.
        :type direct_task: int
        :param need_task_ids: indicate whether or not you'd like the task IDs
                              the Tuple was emitted (default:
                              ``True``).
        :type need_task_ids: bool

        :returns: a ``list`` of task IDs that the Tuple was sent to. Note that
                  when specifying direct_task, this will be equal to
                  ``[direct_task]``. If you specify ``need_task_ids=False``,
                  this function will return ``None``.
        """
        return super(Spout, self).emit(tup, tup_id=tup_id, stream=stream,
                                       direct_task=direct_task,
                                       need_task_ids=need_task_ids)

    def emit_many(self, tuples, stream=None, tup_ids=None, direct_task=None,
                  need_task_ids=True):
        """Emit multiple tuples.

        :param tuples: a ``list`` of multiple Tuple payloads to send to
                       Storm. All Tuples should contain only
                       JSON-serializable data.
        :type tuples: list
        :param stream: the ID of the stream to emit these Tuples to. Specify
                       ``None`` to emit to default stream.
        :type stream: str
        :param tup_ids: the ID for the Tuple. Leave this blank for an
                       unreliable emit.
        :type tup_ids: list
        :param tup_ids: IDs for each of the Tuples in the list.  Omit these for
                        an unreliable emit.
        :type anchors: list
        :param direct_task: indicates the task to send the Tuple to.
        :type direct_task: int
        :param need_task_ids: indicate whether or not you'd like the task IDs
                              the Tuple was emitted (default:
                              ``True``).
        :type need_task_ids: bool

        .. deprecated:: 2.0.0
            Just call :py:meth:`Spout.emit` repeatedly instead.
        """
        if not isinstance(tuples, (list, tuple)):
            raise TypeError('Tuples should be a list of lists/tuples, '
                            'received {!r} instead.'.format(type(tuples)))

        all_task_ids = []
        if tup_ids is None:
            tup_ids = itertools.repeat(None)

        for tup, tup_id in zip(tuples, tup_ids):
            all_task_ids.append(self.emit(tup, stream=stream, tup_id=tup_id,
                                          direct_task=direct_task,
                                          need_task_ids=need_task_ids))

        return all_task_ids

    def _run(self):
        """The inside of ``run``'s infinite loop.

        Separated out so it can be properly unit tested.
        """
        cmd = self.read_command()
        if cmd['command'] == 'next':
            self.next_tuple()
        elif cmd['command'] == 'ack':
            self.ack(cmd['id'])
        elif cmd['command'] == 'fail':
            self.fail(cmd['id'])
        else:
            self.logger.error('Received invalid command from Storm: %r', cmd)
        self.send_message({'command': 'sync'})



from __future__ import print_function, unicode_literals

import subprocess
import time
from contextlib import contextmanager
from socket import error as SocketError

from six.moves.socketserver import UDPServer, TCPServer


def _port_in_use(port, server_type="tcp"):
    """Check to see whether a given port is already in use on localhost."""
    if server_type == "tcp":
        server = TCPServer
    elif server_type == "udp":
        server = UDPServer
    else:
        raise ValueError("Server type can only be: udp or tcp.")

    try:
        server(("localhost", port), None)
    except SocketError:
        return True

    return False


@contextmanager
def ssh_tunnel(user, host, local_port, remote_port):
    if _port_in_use(local_port):
        raise IOError("Local port: {} already in use, unable to open ssh "
                      "tunnel to {}:{}.".format(local_port, host, remote_port))

    if user:
        user_at_host = "{user}@{host}".format(user=user, host=host)
    else:
        user_at_host = host # Rely on SSH default or config to connect.

    ssh_cmd = ["ssh",
               "-NL",
               "{local}:localhost:{remote}".format(
                   local=local_port,
                   remote=remote_port),
               user_at_host]
    ssh_proc = subprocess.Popen(ssh_cmd, shell=False)
    # Validate that the tunnel is actually running before yielding
    while not _port_in_use(local_port):
        # Periodically check to see if the ssh command failed and returned a
        # value, then raise an Exception
        if ssh_proc.poll() is not None:
            raise IOError('Unable to open ssh tunnel via: "{}"'
                          .format(" ".join(ssh_cmd)))
        time.sleep(0.2)
    try:
        yield
    finally:
        ssh_proc.kill()



"""
Create a graphviz visualization of the specified topology.

Unfortunately, this does not currently work, because of packaging difficulities
with Clojure.  In version 2.0, this will replaced with a pure-Python approach to
visualizing topologies.  See GitHub issue #22.
"""

from __future__ import absolute_import, print_function

import sys

from invoke import run

from .common import add_name
from ..util import get_topology_definition


def visualize_topology(name=None, flip=False):
    name, topology_file = get_topology_definition(name)
    print("Visualizing {} topology...".format(name))
    sys.stdout.flush()
    cmd = ["lein",
           "run -m streamparse.commands.visualize/-main",
           topology_file]
    if flip:
        cmd.append("-f")
    full_cmd = " ".join(cmd)
    print("Running lein command to visualize topology:")
    print(full_cmd)
    sys.stdout.flush()
    run(full_cmd)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('visualize',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    subparser.add_argument('-f', '--flip',
                           action='store_true',
                           help='Flip the visualization to be horizontal.')
    add_name(subparser)

def main(args):
    """ Create a graphviz visualization of a topology. """
    visualize_topology(args.name)



"""
Topology base class
"""
from six import add_metaclass, iteritems

from .component import Specification


class Grouping(object):
    SHUFFLE = ":shuffle"
    GLOBAL = ":global"
    DIRECT = ":direct"
    ALL = ":all"

    @classmethod
    def fields(cls, *fieldlist):
        return list(fieldlist)

    @classmethod
    def valid(cls, grouping):
        return (
            isinstance(grouping, list) or
            grouping in (
                Grouping.SHUFFLE,
                Grouping.GLOBAL,
                Grouping.DIRECT,
                Grouping.ALL,
            )
        )


class TopologyType(type):
    def __new__(meta, classname, bases, class_dict):
        specs = {}
        for name, spec in iteritems(class_dict):
            if isinstance(spec, Specification):

                # Use the variable name as the specification name.
                if spec.name is None:
                    spec.name = name

                if spec.name not in specs:
                    specs[spec.name] = spec
                else:
                    raise TopologyError(
                        "Duplicate specification name: {}".format(spec.name))

        class_dict["specs"] = specs.values()

        # Resolve dependencies in specifications.
        for name, spec in iteritems(specs):
            spec.resolve_dependencies(specs)

        return type.__new__(meta, classname, bases, class_dict)


@add_metaclass(TopologyType)
class Topology(object):
    pass


class TopologyError(Exception):
    pass



"""
Kill the specified Storm topology.
"""

from __future__ import absolute_import

from invoke import run

from ..contextmanagers import ssh_tunnel
from ..util import (get_topology_definition, get_env_config,
                    get_nimbus_for_env_config, is_ssh_for_nimbus)
from .common import add_environment, add_name, add_wait


def _kill_topology(topology_name, wait=None,
                   host=None, port=None,
                   run_args=None, run_kwargs=None):
    if run_args is None:
        run_args = []
    if run_kwargs is None:
        run_kwargs = {}
    run_kwargs['pty'] = True
    wait_arg = ("--wait {wait}".format(wait=wait)) if wait is not None else ""
    cmd = ("lein run -m streamparse.commands.kill_topology/-main"
           " {topology_name} {wait}").format(topology_name=topology_name,
                                             wait=wait_arg)
    if host:
        cmd += " --host " + host
    if port:
        cmd += " --port " + str(port)
    return run(cmd, *run_args, **run_kwargs)


def kill_topology(topology_name=None, env_name=None, wait=None):
    topology_name = get_topology_definition(topology_name)[0]
    env_name, env_config = get_env_config(env_name)
    host, port = get_nimbus_for_env_config(env_config)

    if is_ssh_for_nimbus(env_config):
        with ssh_tunnel(env_config.get("user"), host, 6627, port):
            return _kill_topology(topology_name, wait)
    return _kill_topology(topology_name, wait, host=host, port=port)


def subparser_hook(subparsers):
    """ Hook to add subparser for this command. """
    subparser = subparsers.add_parser('kill',
                                      description=__doc__,
                                      help=main.__doc__)
    subparser.set_defaults(func=main)
    add_environment(subparser)
    add_name(subparser)
    add_wait(subparser)


def main(args):
    """ Kill the specified Storm topology """
    kill_topology(topology_name=args.name, env_name=args.environment,
                  wait=args.wait)



"""
Functions for adding common CLI arguments to argparse sub-commands.
"""

def add_ackers(parser):
    """ Add --ackers option to parser """
    parser.add_argument('-a', '--ackers',
                        help='Set number of acker bolts. Takes precedence over '
                             '--par if both set.')


def add_debug(parser):
    """ Add --debug option to parser """
    parser.add_argument('-d', '--debug',
                        action='store_true',
                        help='Set topology.debug and produce debugging output.')


def add_environment(parser):
    """ Add --environment option to parser """
    parser.add_argument('-e', '--environment',
                        help='The environment to use for the command.  '
                             'Corresponds to an environment in your '
                             '"envs" dictionary in config.json.  If you '
                             'only have one environment specified, '
                             'streamparse will automatically use this.')


def add_name(parser):
    """ Add --name option to parser """
    parser.add_argument('-n', '--name',
                        help='The name of the topology to act on.  If you have '
                             'only one topology defined in your "topologies" '
                             'directory, streamparse will use it '
                             'automatically.')


def add_options(parser):
    """ Add --option options to parser """
    parser.add_argument('-o', '--option',
                        dest='options',
                        action='append',
                        help='Topology option to use upon submit.  For example,'
                             ' "-o topology.debug=true" is equivalent to '
                             '"--debug".  May be repeated multiple for multiple'
                             ' options.')


def add_par(parser):
    """ Add --par option to parser """
    parser.add_argument('-p', '--par',
                        default=2,
                        type=int,
                        help='Parallelism of topology; conveniently sets '
                             'number of Storm workers and acker bolts at once '
                             'to passed value. (default: %(default)s)')

def add_pattern(parser):
    """ Add --pattern option to parser """
    parser.add_argument('--pattern',
                        help='Pattern of log files to operate on.')


def add_simple_jar(parser):
    """ Add --simple_jar option to parser. """
    parser.add_argument("-s", "--simple_jar",
                        action='store_true',
                        help='Instead of creating an Uber-JAR for the '
                             'topology, which contains all of its JVM '
                             'dependencies, create a simple JAR with just the '
                             'code for the project.  This is useful when your '
                             'project is pure Python and has no JVM '
                             'dependencies.')


def add_wait(parser):
    """ Add --wait option to parser """
    parser.add_argument('--wait',
                        type=int,
                        default=5,
                        help='Seconds to wait before killing topology. '
                             '(default: %(default)s)')


def add_workers(parser):
    """ Add --workers option to parser """
    parser.add_argument('-w', '--workers',
                        type=int,
                        help='Set number of Storm workers. Takes precedence '
                             'over --par if both set.')


def resolve_ackers_workers(args):
    """ Set --ackers and --workers to --par if they're None. """
    if args.ackers is None:
        args.ackers = args.par
    if args.workers is None:
        args.workers = args.par



from __future__ import absolute_import, print_function, unicode_literals

import os
import re
import shutil
import sys
from glob import glob
from random import shuffle

import requests
import simplejson as json
from fabric.api import env
from fabric.colors import red
from invoke import run
from pkg_resources import parse_version

from .contextmanagers import ssh_tunnel
from .decorators import memoized


def activate_env(env_name=None):
    """Activate a particular environment from a streamparse project's
    config.json file and populatefabric's env dictionary with appropriate
    values.

    :param env_name: a `str` corresponding to the key within the config file's
                     "envs" dictionary.
    """
    env_name, env_config = get_env_config(env_name)

    env.storm_workers = env_config["workers"]
    env.user = env_config.get("user")
    env.log_path = env_config.get("log_path") or \
                   env_config.get("log", {}).get("path")
    env.virtualenv_root = env_config.get("virtualenv_root") or \
                          env_config.get("virtualenv_path")
    env.disable_known_hosts = True
    env.forward_agent = True
    env.use_ssh_config = True


def die(msg, error_code=1):
    print("{}: {}".format(red("error"), msg))
    sys.exit(error_code)


@memoized
def get_config():
    if not os.path.exists("config.json"):
        die("No config.json found. You must run this command inside a "
            "streamparse project directory.")

    with open("config.json") as fp:
        config = json.load(fp)
    return config


def get_topology_definition(topology_name=None):
    """Fetch a topology name and definition file.  If the topology_name is
    None, and there's only one topology definiton listed, we'll select that
    one, otherwise we'll die to avoid ambiguity.

    :param topology_name: a `str`, the topology_name of the topology (without
                          .clj extension).
    :returns: a `tuple` containing (topology_topology_name, topology_file).
    """
    config = get_config()
    topology_path = config["topology_specs"]
    if topology_name is None:
        topology_files = glob("{}/*.clj".format(topology_path))
        if not topology_files:
            die("No topology definitions are defined in {}."
                .format(topology_path))
        if len(topology_files) > 1:
            die("Found more than one topology definition file in {specs_dir}. "
                "When more than one topology definition file exists, you must "
                "explicitly specify the topology by name using the -n or "
                "--name flags.".format(specs_dir=topology_path))
        topology_file = topology_files[0]
        topology_name = re.sub(r'(^{}|\.clj$)'.format(topology_path), '',
                               topology_file)
    else:
        topology_file = "{}.clj".format(os.path.join(topology_path,
                                                     topology_name))
        if not os.path.exists(topology_file):
            die("Topology definition file not found {}. You need to "
                "create a topology definition file first."
                .format(topology_file))

    return (topology_name, topology_file)


def get_env_config(env_name=None):
    """Fetch an environment name and config object from the config.json file.
    If the name is None and there's only one environment, we'll select the
    first, otherwise we'll die to avoid ambiguity.

    :returns: a `tuple` containing (env_name, env_config).
    """
    config = get_config()
    if env_name is None and len(config["envs"]) == 1:
        env_name = list(config["envs"].keys())[0]
    elif env_name is None and len(config["envs"]) > 1:
        die("Found more than one environment in config.json. "
            "When more than one environment exists, you must "
            "explicitly specify the environment name via the "
            "-e or --environment flags.")
    if env_name not in config["envs"]:
        die("Could not find a \"{}\" in config.json, have you specified one?"
            .format(env_name))

    return (env_name, config["envs"][env_name])


def get_nimbus_for_env_config(env_config):
    """Get the Nimbus server's hostname and port from a streamparse project's
    config file.
    """
    if not env_config["nimbus"]:
        die("No Nimbus server configured in config.json.")

    if ":" in env_config["nimbus"]:
        host, port = env_config["nimbus"].split(":", 1)
        port = int(port)
    else:
        host = env_config["nimbus"]
        port = 6627

    return (host, port)


def is_ssh_for_nimbus(env_config):
    """Check if we need to use SSH access to Nimbus or not.
    """
    return env_config.get('use_ssh_for_nimbus', True)


def storm_lib_version():
    """Get the Storm library version being used by Leiningen and Streamparse.

    :returns: The Storm library version specified in project.clj
    :rtype: pkg_resources.Version
    """
    deps_tree = run("lein deps :tree", hide=True).stdout
    pattern = r'\[org\.apache\.storm/storm-core "([^"]+)"\]'
    versions = set(re.findall(pattern, deps_tree))

    if len(versions) > 1:
        raise RuntimeError("Multiple Storm Versions Detected.")
    elif len(versions) == 0:
        raise RuntimeError("No Storm version specified in project.clj "
                           "dependencies.")
    else:
        return parse_version(versions.pop())


def get_ui_jsons(env_name, api_paths):
    """Take env_name as a string and api_paths that should
    be a list of strings like '/api/v1/topology/summary'
    """
    _, env_config = get_env_config(env_name)
    host, _ = get_nimbus_for_env_config(env_config)
    # TODO: Get remote_ui_port from storm?
    remote_ui_port = 8080
    # SSH tunnel can take a while to close. Check multiples if necessary.
    local_ports = list(range(8081, 8090))
    shuffle(local_ports)
    for local_port in local_ports:
        try:
            data = {}
            with ssh_tunnel(env_config.get("user"), host, local_port=local_port,
                            remote_port=remote_ui_port):
                for api_path in api_paths:
                    r = requests.get('http://127.0.0.1:%s%s' % (local_port,
                                                                api_path))
                    data[api_path] = r.json()
            return data
        except Exception as e:
            if "already in use" in str(e):
                continue
            raise
    raise RuntimeError("Cannot find local port for SSH tunnel to Storm Head.")


def get_ui_json(env_name, api_path):
    """Take env_name as a string and api_path that should
    be a string like '/api/v1/topology/summary'
    """
    return get_ui_jsons(env_name, [api_path])[api_path]


def prepare_topology():
    """Prepare a topology for running locally or deployment to a remote
    cluster.
    """
    if os.path.isdir("_resources/resources"):
        shutil.rmtree("_resources/resources")
    shutil.copytree("src", "_resources/resources")


def _get_file_names_command(path, patterns):
    """Given a list of bash `find` patterns, return a string for the
    bash command that will find those streamparse log files
    """
    patterns = "' -o -name '".join(patterns)
    return ("cd {path} && "
            "find . -maxdepth 1 -name '{patterns}' ") \
            .format(path=path, patterns=patterns)


def get_logfiles_cmd(topology_name=None, pattern=None):
    """ Returns a string representing a command to run on the Storm workers that
    will yield all of the logfiles for the given topology that meet the given
    pattern (if specified).
    """
    # list log files found
    log_name_patterns = ["worker*", "supervisor*", "access*", "metrics*",
                         "streamparse_{topo_name}*".format(topo_name=topology_name)]
    ls_cmd = _get_file_names_command(env.log_path, log_name_patterns)
    if pattern is not None:
        ls_cmd += " | egrep '{pattern}'".format(pattern=pattern)
    return ls_cmd
